@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -117,4 +117,15 @@ typedef enum {
     _NOREG,
 } jit_reg_t;
 
+typedef struct {
+    jit_uint32_t clz		: 1;
+    /* generate special instructions for unaligned load/store? */
+    jit_uint32_t unaligned	: 1;
+} jit_cpu_t;
+
+/*
+ * Initialization
+ */
+extern jit_cpu_t		jit_cpu;
+
 #endif /* _jit_ia64_h */
