@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -250,8 +250,48 @@ extern jit_node_t *_jit_data(jit_state_t*, const void*,
     } while (0)
 #define jit_inc_synth_dp(name, u, v)					\
     jit_code_inc_synth_dp(jit_code_##name, u, v)
+#define jit_inc_synth_wf(name, u, v)					\
+    jit_code_inc_synth_wf(jit_code_##name, u, v)
+#define jit_code_inc_synth_wf(code, u, v)				\
+    do {								\
+	(void)jit_new_node_wf(code, u, v);				\
+	jit_synth_inc();						\
+    } while (0)
+#define jit_inc_synth_wqf(name, u, v, w, x)				\
+    do {								\
+	(void)jit_new_node_wqf(jit_code_##name, u, v, w, x);		\
+	jit_synth_inc();						\
+    } while (0)
+#define jit_inc_synth_wd(name, u, v)					\
+    jit_code_inc_synth_wd(jit_code_##name, u, v)
+#define jit_inc_synth_wqd(name, u, v, w, x)				\
+    do {								\
+	(void)jit_new_node_wqd(jit_code_##name, u, v, w, x);		\
+	jit_synth_inc();						\
+    } while (0)
+#define jit_code_inc_synth_wd(code, u, v)				\
+    do {								\
+	(void)jit_new_node_wd(code, u, v);				\
+	jit_synth_inc();						\
+    } while (0)
 #define jit_dec_synth()		jit_synth_dec()
 
+#define jit_link_alist(node)						\
+    do {								\
+	node->link = _jitc->function->alist;				\
+	_jitc->function->alist = node;					\
+    } while (0)
+#define jit_check_frame()						\
+    do {								\
+	if (!_jitc->function->need_frame) {				\
+	    _jitc->again = 1;						\
+	    _jitc->function->need_frame = 1;				\
+	}								\
+    } while (0)
+#define jit_diffsize()	(stack_framesize - _jitc->framesize)
+#define jit_framesize()	(stack_framesize - jit_diffsize())
+#define jit_selfsize()	(_jitc->function->self.size - jit_diffsize())
+
 #define jit_link_prolog()						\
     do {								\
 	_jitc->tail->link = _jitc->function->prolog->link;		\
@@ -310,10 +350,11 @@ extern jit_node_t *_jit_data(jit_state_t*, const void*,
 #define jit_cc_a0_cnd		0x00000100	/* arg1 is a conditinally set register */
 #define jit_cc_a1_reg		0x00000200	/* arg1 is a register */
 #define jit_cc_a1_chg		0x00000400	/* arg1 is modified */
-#define jit_cc_a1_int		0x00001000	/* arg1 is immediate word */
-#define jit_cc_a1_flt		0x00002000	/* arg1 is immediate float */
-#define jit_cc_a1_dbl		0x00004000	/* arg1 is immediate double */
-#define jit_cc_a1_arg		0x00008000	/* arg1 is an argument node */
+#define jit_cc_a1_int		0x00000800	/* arg1 is immediate word */
+#define jit_cc_a1_flt		0x00001000	/* arg1 is immediate float */
+#define jit_cc_a1_dbl		0x00002000	/* arg1 is immediate double */
+#define jit_cc_a1_arg		0x00004000	/* arg1 is an argument node */
+#define jit_cc_a1_rlh		0x00008000	/* arg1 is a register pair */
 #define jit_cc_a2_reg		0x00010000	/* arg2 is a register */
 #define jit_cc_a2_chg		0x00020000	/* arg2 is modified */
 #define jit_cc_a2_int		0x00100000	/* arg2 is immediate word */
@@ -510,9 +551,17 @@ struct jit_function {
     } call;
     jit_node_t		*prolog;
     jit_node_t		*epilog;
+    jit_node_t		*alist;
     jit_int32_t		*regoff;
     jit_regset_t	 regset;
     jit_int32_t		 stack;
+#if defined(__i386__) || defined(__x86_64__) || \
+    defined(__powerpc__) || defined(__sparc__) || \
+    defined(__s390__) || defined(__s390x__) || \
+    defined(__hppa__) || defined(__alpha__)
+    jit_int32_t		 cvt_offset;	/* allocai'd offset for x87<->xmm or
+					 * fpr<->gpr transfer using the stack */
+#endif
 
     /* Helper for common jit generation pattern, used in GNU Smalltalk
      * and possibly others, where a static frame layout is required or
@@ -521,11 +570,25 @@ struct jit_function {
     jit_uint32_t	 define_frame : 1;
     jit_uint32_t	 assume_frame : 1;
 
+    jit_uint32_t	 need_frame : 1;	/* need frame pointer? */
+    jit_uint32_t	 need_stack : 1;	/* need stack pointer? */
+    jit_uint32_t	 need_return : 1;	/* not a leaf function */
+
     /* alloca offset offset */
     jit_int32_t		 aoffoff;
     /* uses allocar flag */
     jit_uint32_t	 allocar : 1;
 
+#if __arm__
+    /* If will, or might use float registers and vfp is not available.
+     * Use the first 64 bytes always, as the access to the virtual float
+     * registers use hardcoded instructions that can only reach 64 byte
+     * displacements, and to keep code simpler, do not use temporaries. */
+    jit_uint32_t	  swf_offset : 1;
+    /* If need to call C functions for some operation, or variadic function */
+    jit_uint32_t	  save_reg_args : 1;
+#endif
+
     /* varargs state offsets */
     jit_int32_t		 vaoff;		/* offset of jit_va_list */
     jit_int32_t		 vagp;		/* first gp va argument */
@@ -545,6 +608,13 @@ struct jit_compiler {
     jit_int32_t		  rout;		/* first output register */
     jit_int32_t		  breg;		/* base register for prolog/epilog */
 #endif
+#if __mips__
+    struct {
+	jit_int32_t	  op;		/* pending instruction, candidate
+					 * to be inserted in a delay slot */
+	jit_bool_t	  pend;		/* non zero if need to emit op */
+    } inst;
+#endif
 #if __mips__ || __ia64__ || __alpha__ || \
 	(__sparc__ && __WORDSIZE == 64) || __riscv || __loongarch__
     jit_int32_t		  carry;
@@ -564,6 +634,14 @@ struct jit_compiler {
 #endif
     jit_uint32_t	  no_data : 1;
     jit_uint32_t	  no_note : 1;
+    /* FIXME undocumented, might be moved to a jit_cpu field or a better
+     * configuration api.
+     * These are switches to a different unld* or unst*.
+     * Defaults are the algorithms that generate shorter code*/
+    jit_uint32_t	  unld_algorithm : 1;
+    jit_uint32_t	  unst_algorithm : 1;
+    jit_int32_t		  framesize;	/* space for callee save registers,
+					 * frame pointer and return address */
     jit_int32_t		  reglen;	/* number of registers */
     jit_regset_t	  regarg;	/* cannot allocate */
     jit_regset_t	  regsav;	/* automatic spill only once */
