@@ -4,7 +4,7 @@
 #define JIT_INSTR_MAX 96
     0,	/* data */
     0,	/* live */
-    12,	/* align */
+    4,	/* align */
     0,	/* save */
     0,	/* load */
     4,	/* skip */
@@ -77,11 +77,11 @@
     12,	/* remr_u */
     20,	/* remi_u */
     4,	/* andr */
-    20,	/* andi */
+    4,	/* andi */
     4,	/* orr */
-    20,	/* ori */
+    4,	/* ori */
     4,	/* xorr */
-    20,	/* xori */
+    4,	/* xori */
     4,	/* lshr */
     4,	/* lshi */
     4,	/* rshr */
@@ -89,7 +89,9 @@
     4,	/* rshr_u */
     4,	/* rshi_u */
     4,	/* negr */
+    4,	/* negi */
     4,	/* comr */
+    4,	/* comi */
     8,	/* ltr */
     8,	/* lti */
     8,	/* ltr_u */
@@ -117,17 +119,29 @@
     28,	/* casr */
     36,	/* casi */
     4,	/* extr_c */
+    4,	/* exti_c */
     4,	/* extr_uc */
+    4,	/* exti_uc */
     4,	/* extr_s */
+    4,	/* exti_s */
     4,	/* extr_us */
+    4,	/* exti_us */
     4,	/* extr_i */
+    12,	/* exti_i */
     4,	/* extr_ui */
+    4,	/* exti_ui */
     8,	/* bswapr_us */
+    4,	/* bswapi_us */
     8,	/* bswapr_ui */
+    8,	/* bswapi_ui */
     4,	/* bswapr_ul */
+    16,	/* bswapi_ul */
     8,	/* htonr_us */
+    4,	/* htoni_us */
     8,	/* htonr_ui */
+    8,	/* htoni_ui */
     4,	/* htonr_ul */
+    16,	/* htoni_ul */
     4,	/* ldr_c */
     16,	/* ldi_c */
     4,	/* ldr_uc */
@@ -270,8 +284,11 @@
     4,	/* divr_f */
     12,	/* divi_f */
     4,	/* negr_f */
+    0,	/* negi_f */
     4,	/* absr_f */
+    0,	/* absi_f */
     4,	/* sqrtr_f */
+    0,	/* sqrti_f */
     8,	/* ltr_f */
     16,	/* lti_f */
     8,	/* ler_f */
@@ -361,8 +378,11 @@
     4,	/* divr_d */
     12,	/* divi_d */
     4,	/* negr_d */
+    0,	/* negi_d */
     4,	/* absr_d */
+    0,	/* absi_d */
     4,	/* sqrtr_d */
+    0,	/* sqrti_d */
     8,	/* ltr_d */
     16,	/* lti_d */
     8,	/* ler_d */
@@ -438,18 +458,76 @@
     0,	/* retr_d */
     0,	/* reti_d */
     0,	/* retval_d */
-    0,	/* movr_w_f */
+    4,	/* movr_w_f */
+    8,	/* movi_w_f */
     0,	/* movr_ww_d */
-    0,	/* movr_w_d */
-    0,	/* movr_f_w */
-    0,	/* movi_f_w */
+    0,	/* movi_ww_d */
+    4,	/* movr_w_d */
+    8,	/* movi_w_d */
+    4,	/* movr_f_w */
+    4,	/* movi_f_w */
     0,	/* movr_d_ww */
     0,	/* movi_d_ww */
-    0,	/* movr_d_w */
-    0,	/* movi_d_w */
-
+    4,	/* movr_d_w */
+    4,	/* movi_d_w */
+    8,	/* clor */
+    4,	/* cloi */
+    4,	/* clzr */
+    4,	/* clzi */
+    12,	/* ctor */
+    4,	/* ctoi */
+    8,	/* ctzr */
+    4,	/* ctzi */
+    4,	/* rbitr */
+    16,	/* rbiti */
+    16,	/* popcntr */
+    4,	/* popcnti */
+    12,	/* lrotr */
+    4,	/* lroti */
+    4,	/* rrotr */
+    4,	/* rroti */
+    4,	/* extr */
+    4,	/* exti */
+    4,	/* extr_u */
+    4,	/* exti_u */
+    4,	/* depr */
+    8,	/* depi */
+    56,	/* qlshr */
+    8,	/* qlshi */
+    52,	/* qlshr_u */
+    8,	/* qlshi_u */
+    52,	/* qrshr */
+    8,	/* qrshi */
+    52,	/* qrshr_u */
+    8,	/* qrshi_u */
+    24,	/* unldr */
+    44,	/* unldi */
+    24,	/* unldr_u */
+    44,	/* unldi_u */
+    20,	/* unstr */
+    56,	/* unsti */
+    8,	/* unldr_x */
+    20,	/* unldi_x */
+    8,	/* unstr_x */
+    20,	/* unsti_x */
+    4,	/* fmar_f */
+    0,	/* fmai_f */
+    4,	/* fmsr_f */
+    0,	/* fmsi_f */
+    4,	/* fmar_d */
+    0,	/* fmai_d */
+    4,	/* fmsr_d */
+    0,	/* fmsi_d */
+    4,	/* fnmar_f */
+    0,	/* fnmai_f */
+    4,	/* fnmsr_f */
+    0,	/* fnmsi_f */
+    4,	/* fnmar_d */
+    0,	/* fnmai_d */
+    4,	/* fnmsr_d */
+    0,	/* fnmsi_d */
 #  else	/* PACKED_STACK */
-#define JIT_INSTR_MAX 96
+#define JIT_INSTR_MAX 120
     0,	/* data */
     0,	/* live */
     12,	/* align */
@@ -459,7 +537,7 @@
     0,	/* #name */
     0,	/* #note */
     0,	/* label */
-    96,	/* prolog */
+    120,	/* prolog */
     0,	/* ellipsis */
     0,	/* va_push */
     0,	/* allocai */
@@ -489,9 +567,9 @@
     0,	/* putargi_ui */
     0,	/* putargr_l */
     0,	/* putargi_l */
-    4,	/* va_start */
-    8,	/* va_arg */
-    12,	/* va_arg_d */
+    44,	/* va_start */
+    48,	/* va_arg */
+    56,	/* va_arg_d */
     0,	/* va_end */
     4,	/* addr */
     20,	/* addi */
@@ -525,11 +603,11 @@
     12,	/* remr_u */
     20,	/* remi_u */
     4,	/* andr */
-    20,	/* andi */
+    4,	/* andi */
     4,	/* orr */
-    20,	/* ori */
+    4,	/* ori */
     4,	/* xorr */
-    20,	/* xori */
+    4,	/* xori */
     4,	/* lshr */
     4,	/* lshi */
     4,	/* rshr */
@@ -537,7 +615,9 @@
     4,	/* rshr_u */
     4,	/* rshi_u */
     4,	/* negr */
+    4,	/* negi */
     4,	/* comr */
+    4,	/* comi */
     8,	/* ltr */
     8,	/* lti */
     8,	/* ltr_u */
@@ -563,19 +643,31 @@
     8,	/* movnr */
     8,	/* movzr */
     28,	/* casr */
-    36,	/* casi */
+    40,	/* casi */
     4,	/* extr_c */
+    4,	/* exti_c */
     4,	/* extr_uc */
+    4,	/* exti_uc */
     4,	/* extr_s */
+    4,	/* exti_s */
     4,	/* extr_us */
+    4,	/* exti_us */
     4,	/* extr_i */
+    12,	/* exti_i */
     4,	/* extr_ui */
+    4,	/* exti_ui */
     8,	/* bswapr_us */
+    4,	/* bswapi_us */
     8,	/* bswapr_ui */
+    8,	/* bswapi_ui */
     4,	/* bswapr_ul */
+    16,	/* bswapi_ul */
     8,	/* htonr_us */
+    4,	/* htoni_us */
     8,	/* htonr_ui */
+    8,	/* htoni_ui */
     4,	/* htonr_ul */
+    16,	/* htoni_ul */
     4,	/* ldr_c */
     16,	/* ldi_c */
     4,	/* ldr_uc */
@@ -718,8 +810,11 @@
     4,	/* divr_f */
     12,	/* divi_f */
     4,	/* negr_f */
+    0,	/* negi_f */
     4,	/* absr_f */
+    0,	/* absi_f */
     4,	/* sqrtr_f */
+    0,	/* sqrti_f */
     8,	/* ltr_f */
     16,	/* lti_f */
     8,	/* ler_f */
@@ -809,8 +904,11 @@
     4,	/* divr_d */
     12,	/* divi_d */
     4,	/* negr_d */
+    0,	/* negi_d */
     4,	/* absr_d */
+    0,	/* absi_d */
     4,	/* sqrtr_d */
+    0,	/* sqrti_d */
     8,	/* ltr_d */
     16,	/* lti_d */
     8,	/* ler_d */
@@ -886,14 +984,73 @@
     0,	/* retr_d */
     0,	/* reti_d */
     0,	/* retval_d */
-    0,	/* movr_w_f */
+    4,	/* movr_w_f */
+    8,	/* movi_w_f */
     0,	/* movr_ww_d */
-    0,	/* movr_w_d */
-    0,	/* movr_f_w */
-    0,	/* movi_f_w */
+    0,	/* movi_ww_d */
+    4,	/* movr_w_d */
+    8,	/* movi_w_d */
+    4,	/* movr_f_w */
+    4,	/* movi_f_w */
     0,	/* movr_d_ww */
     0,	/* movi_d_ww */
-    0,	/* movr_d_w */
-    0,	/* movi_d_w */
+    4,	/* movr_d_w */
+    4,	/* movi_d_w */
+    8,	/* clor */
+    4,	/* cloi */
+    4,	/* clzr */
+    4,	/* clzi */
+    12,	/* ctor */
+    4,	/* ctoi */
+    8,	/* ctzr */
+    4,	/* ctzi */
+    4,	/* rbitr */
+    16,	/* rbiti */
+    16,	/* popcntr */
+    4,	/* popcnti */
+    12,	/* lrotr */
+    4,	/* lroti */
+    4,	/* rrotr */
+    4,	/* rroti */
+    4,	/* extr */
+    4,	/* exti */
+    4,	/* extr_u */
+    4,	/* exti_u */
+    4,	/* depr */
+    8,	/* depi */
+    56,	/* qlshr */
+    8,	/* qlshi */
+    52,	/* qlshr_u */
+    8,	/* qlshi_u */
+    52,	/* qrshr */
+    8,	/* qrshi */
+    52,	/* qrshr_u */
+    8,	/* qrshi_u */
+    24,	/* unldr */
+    44,	/* unldi */
+    24,	/* unldr_u */
+    44,	/* unldi_u */
+    20,	/* unstr */
+    56,	/* unsti */
+    8,	/* unldr_x */
+    20,	/* unldi_x */
+    8,	/* unstr_x */
+    20,	/* unsti_x */
+    4,	/* fmar_f */
+    0,	/* fmai_f */
+    4,	/* fmsr_f */
+    0,	/* fmsi_f */
+    4,	/* fmar_d */
+    0,	/* fmai_d */
+    4,	/* fmsr_d */
+    0,	/* fmsi_d */
+    4,	/* fnmar_f */
+    0,	/* fnmai_f */
+    4,	/* fnmsr_f */
+    0,	/* fnmsi_f */
+    4,	/* fnmar_d */
+    0,	/* fnmai_d */
+    4,	/* fnmsr_d */
+    0,	/* fnmsi_d */
 #  endif
 #endif /* __WORDSIZE */
