@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2014-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -27,10 +27,8 @@ static char *code_name[] = {
     "prolog",
     "ellipsis",		"va_push",
     "allocai",		"allocar",
-    "arg_c",
-    "arg_s",
-    "arg_i",
-    "arg_l",
+    "arg_c",		"arg_s",
+    "arg_i",		"arg_l",
     "getarg_c",		"getarg_uc",
     "getarg_s",		"getarg_us",
     "getarg_i",		"getarg_ui",
@@ -67,7 +65,8 @@ static char *code_name[] = {
     "lshr",		"lshi",
     "rshr",		"rshi",
     "rshr_u",		"rshi_u",
-    "negr",		"comr",
+    "negr",		"negi",
+    "comr",		"comi",
     "ltr",		"lti",
     "ltr_u",		"lti_u",
     "ler",		"lei",
@@ -81,13 +80,18 @@ static char *code_name[] = {
     "movr",		"movi",
     "movnr",		"movzr",
     "casr",		"casi",
-    "extr_c",		"extr_uc",
-    "extr_s",		"extr_us",
-    "extr_i",		"extr_ui",
-    "bswapr_us",
-    "bswapr_ui",	"bswapr_ul",
-    "htonr_us",
-    "htonr_ui",		"htonr_ul",
+    "extr_c",		"exti_c",
+    "extr_uc",		"exti_uc",
+    "extr_s",		"exti_s",
+    "extr_us",		"exti_us",
+    "extr_i",		"exti_i",
+    "extr_ui",		"exti_ui",
+    "bswapr_us",	"bswapi_us",
+    "bswapr_ui",	"bswapi_ui",
+    "bswapr_ul",	"bswapi_ul",
+    "htonr_us",		"htoni_us",
+    "htonr_ui",		"htoni_ui",
+    "htonr_ul",		"htoni_ul",
     "ldr_c",		"ldi_c",
     "ldr_uc",		"ldi_uc",
     "ldr_s",		"ldi_s",
@@ -161,8 +165,9 @@ static char *code_name[] = {
     "rsbi_f",
     "mulr_f",		"muli_f",
     "divr_f",		"divi_f",
-    "negr_f",		"absr_f",
-    "sqrtr_f",
+    "negr_f",		"negi_f",
+    "absr_f",		"absi_f",
+    "sqrtr_f",		"sqrti_f",
     "ltr_f",		"lti_f",
     "ler_f",		"lei_f",
     "eqr_f",		"eqi_f",
@@ -208,8 +213,9 @@ static char *code_name[] = {
     "rsbi_d",
     "mulr_d",		"muli_d",
     "divr_d",		"divi_d",
-    "negr_d",		"absr_d",
-    "sqrtr_d",
+    "negr_d",		"negi_d",
+    "absr_d",		"absi_d",
+    "sqrtr_d",		"sqrti_d",
     "ltr_d",		"lti_d",
     "ler_d",		"lei_d",
     "eqr_d",		"eqi_d",
@@ -224,7 +230,8 @@ static char *code_name[] = {
     "ltgtr_d",		"ltgti_d",
     "ordr_d",		"ordi_d",
     "unordr_d",		"unordi_d",
-    "truncr_d_i",	"truncr_d_l",
+    "truncr_d_i",
+    "truncr_d_l",
     "extr_d",		"extr_f_d",
     "movr_d",		"movi_d",
     "ldr_d",		"ldi_d",
@@ -248,9 +255,38 @@ static char *code_name[] = {
     "pushargr_d",	"pushargi_d",
     "retr_d",		"reti_d",
     "retval_d",
-    "movr_w_f",		"movr_ww_d",
-    "movr_w_d",
+    "movr_w_f",		"movi_w_f",
+    "movr_ww_d",	"movi_ww_d",
+    "movr_w_d",		"movi_w_d",
     "movr_f_w",		"movi_f_w",
     "movr_d_ww",	"movi_d_ww",
     "movr_d_w",		"movi_d_w",
+    "clor",		"cloi",
+    "clzr",		"clzi",
+    "ctor",		"ctoi",
+    "ctzr",		"ctzi",
+    "rbitr",		"rbiti",
+    "popcntr",		"popcnti",
+    "lrotr",		"lroti",
+    "rrotr",		"rroti",
+    "extr",		"exti",
+    "extr_u",		"exti_u",
+    "depr",		"depi",
+    "qlshr",		"qlshi",
+    "qlshr_u",		"qlshi_u",
+    "qrshr",		"qrshi",
+    "qrshr_u",		"qrshi_u",
+    "unldr",		"unldi",
+    "unldr_u",		"unldi_u",
+    "unstr",		"unsti",
+    "unldr_x",		"unldi_x",
+    "unstr_x",		"unsti_x",
+    "fmar_f",		"fmai_f",
+    "fmsr_f",		"fmsi_f",
+    "fmar_d",		"fmai_d",
+    "fmsr_d",		"fmsi_d",
+    "fnmar_f",		"fnmai_f",
+    "fnmsr_f",		"fnmsi_f",
+    "fnmar_d",		"fnmai_d",
+    "fnmsr_d",		"fnmsi_d",
 };
