@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2019-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2019-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -17,6 +17,10 @@
  *	Paulo Cesar Pereira de Andrade
  */
 
+/* callee save					  + variadic arguments
+ * align16(ra+fp+s[1-9]+s10+s11+fs[0-9]+fs10+fs11)+align16(a[0-7]) */
+#define stack_framesize			(208 + 64)
+
 #define jit_arg_reg_p(i)		((i) >= 0 && (i) < 8)
 #define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 8)
 
@@ -28,6 +32,8 @@ typedef jit_pointer_t jit_va_list_t;
 /*
  * Prototypes
  */
+#define compute_framesize()		_compute_framesize(_jit)
+static void _compute_framesize(jit_state_t*);
 #if __WORDSIZE == 64
 #  define load_const(r0, i0)		_load_const(_jit, r0, i0)
 static void _load_const(jit_state_t*, jit_int32_t, jit_word_t);
@@ -43,11 +49,13 @@ static void _patch(jit_state_t*,jit_word_t,jit_node_t*);
 #define PROTO				1
 #  include "jit_riscv-cpu.c"
 #  include "jit_riscv-fpu.c"
+#  include "jit_fallback.c"
 #undef PROTO
 
 /*
  * Initialization
  */
+jit_cpu_t		jit_cpu;
 jit_register_t		_rvs[] = {
     { 0x00,				"zero" },
     { 0x01,				"ra" },
@@ -119,12 +127,22 @@ jit_register_t		_rvs[] = {
     { _NOREG,				"<none>" },
 };
 
+static jit_int32_t iregs[] = {
+    _S1, _S2, _S3, _S4, _S5, _S6, _S7, _S8, _S9, _S10, _S11
+};
+
+static jit_int32_t fregs[] = {
+    _FS0, _FS1, _FS2, _FS3, _FS4, _FS5, _FS6, _FS7, _FS8, _FS9, _FS10, _FS11
+};
+
 /*
  * Implementation
  */
 void
 jit_get_cpu(void)
 {
+    /* By default generate extra instructions for unaligned load/store. */
+    jit_cpu.unaligned = 0;
 }
 
 void
@@ -180,6 +198,7 @@ jit_int32_t
 _jit_allocai(jit_state_t *_jit, jit_int32_t length)
 {
     assert(_jitc->function);
+    jit_check_frame();
     switch (length) {
 	case 0:	case 1:						break;
 	case 2:		_jitc->function->self.aoff &= -2;	break;
@@ -302,13 +321,14 @@ _jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
     if (u->code >= jit_code_arg_c && u->code <= jit_code_arg)
 	return (jit_arg_reg_p(u->u.w));
     assert(u->code == jit_code_arg_f || u->code == jit_code_arg_d);
-    return (jit_arg_f_reg_p(u->u.w));
+    return (jit_arg_f_reg_p(u->u.w) || jit_arg_reg_p(u->u.w - 8));
 }
 
 void
 _jit_ellipsis(jit_state_t *_jit)
 {
     jit_inc_synth(ellipsis);
+    jit_check_frame();
     if (_jitc->prepare) {
 	jit_link_prepare();
 	assert(!(_jitc->function->call.call & jit_call_varargs));
@@ -346,6 +366,7 @@ _jit_arg(jit_state_t *_jit, jit_code_t code)
     else {
 	offset = _jitc->function->self.size;
 	_jitc->function->self.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     node = jit_new_node_ww(code, offset,
 			   ++_jitc->function->self.argn);
@@ -369,6 +390,7 @@ _jit_arg_f(jit_state_t *_jit)
     else {
 	offset = _jitc->function->self.size;
 	_jitc->function->self.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     node = jit_new_node_ww(jit_code_arg_f, offset,
 			   ++_jitc->function->self.argn);
@@ -392,6 +414,7 @@ _jit_arg_d(jit_state_t *_jit)
     else {
 	offset = _jitc->function->self.size;
 	_jitc->function->self.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     node = jit_new_node_ww(jit_code_arg_d, offset,
 			   ++_jitc->function->self.argn);
@@ -406,8 +429,10 @@ _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_c, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_c(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_c(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -418,8 +443,10 @@ _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_uc, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_uc(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_uc(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -430,8 +457,10 @@ _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_s, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_s(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_s(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -442,8 +471,10 @@ _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_us, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_us(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_us(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -454,8 +485,10 @@ _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_i, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_i(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_i(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_i(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -466,8 +499,10 @@ _jit_getarg_ui(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_ui, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_extr_ui(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_ui(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_ui(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -478,8 +513,10 @@ _jit_getarg_l(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
     jit_inc_synth_wp(getarg_l, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, JIT_RA0 - v->u.w);
-    else
-	jit_ldxi_l(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_l(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -490,8 +527,10 @@ _jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v, jit_code_t code)
     jit_code_inc_synth_wp(code, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movr(JIT_RA0 - v->u.w, u);
-    else
-	jit_stxi(v->u.w, JIT_FP, u);
+    else {
+	jit_node_t	*node = jit_stxi(v->u.w, JIT_FP, u);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -504,9 +543,11 @@ _jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v, jit_code_t code)
     if (jit_arg_reg_p(v->u.w))
 	jit_movi(JIT_RA0 - v->u.w, u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_gpr);
 	jit_movi(regno, u);
-	jit_stxi(v->u.w, JIT_FP, regno);
+	node = jit_stxi(v->u.w, JIT_FP, regno);
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -521,8 +562,10 @@ _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_movr_f(u, JIT_FA0 - v->u.w);
     else if (jit_arg_reg_p(v->u.w - 8))
 	jit_movr_w_f(u, JIT_RA0 - (v->u.w - 8));
-    else
-	jit_ldxi_f(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_f(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -535,8 +578,10 @@ _jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_movr_f(JIT_FA0 - v->u.w, u);
     else if (jit_arg_reg_p(v->u.w - 8))
 	jit_movr_f_w(JIT_RA0 - (v->u.w - 8), u);
-    else
-	jit_stxi_f(v->u.w, JIT_FP, u);
+    else {
+	jit_node_t	*node = jit_stxi_f(v->u.w, JIT_FP, u);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -548,18 +593,14 @@ _jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
     jit_inc_synth_fp(putargi_f, u, v);
     if (jit_arg_f_reg_p(v->u.w))
 	jit_movi_f(JIT_FA0 - v->u.w, u);
-    else if (jit_arg_reg_p(v->u.w - 8)) {
-	union {
-	    jit_float32_t	f;
-	    jit_int32_t		i;
-	} uu;
-	uu.f = u;
-	jit_movi(JIT_RA0 - (v->u.w - 8), uu.i);
-    }
+    else if (jit_arg_reg_p(v->u.w - 8))
+	jit_movi_f_w(JIT_RA0 - (v->u.w - 8), u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
-	jit_stxi_f(v->u.w, JIT_FP, regno);
+	node = jit_stxi_f(v->u.w, JIT_FP, regno);
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -574,8 +615,10 @@ _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_movr_d(u, JIT_FA0 - v->u.w);
     else if (jit_arg_reg_p(v->u.w - 8))
 	jit_movr_w_d(u, JIT_RA0 - (v->u.w - 8));
-    else
-	jit_ldxi_d(u, JIT_FP, v->u.w);
+    else {
+	jit_node_t	*node = jit_ldxi_d(u, JIT_FP, v->u.w);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -588,8 +631,10 @@ _jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	jit_movr_d(JIT_FA0 - v->u.w, u);
     else if (jit_arg_reg_p(v->u.w - 8))
 	jit_movr_d_w(JIT_RA0 - (v->u.w - 8), u);
-    else
-	jit_stxi_d(v->u.w, JIT_FP, u);
+    else {
+	jit_node_t	*node = jit_stxi_d(v->u.w, JIT_FP, u);
+	jit_link_alist(node);
+    }
     jit_dec_synth();
 }
 
@@ -601,18 +646,14 @@ _jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
     jit_inc_synth_dp(putargi_d, u, v);
     if (jit_arg_reg_p(v->u.w))
 	jit_movi_d(JIT_FA0 - v->u.w, u);
-    else if (jit_arg_reg_p(v->u.w - 8)) {
-	union {
-	    jit_float64_t	d;
-	    jit_int64_t		w;
-	} uu;
-	uu.d = u;
-	jit_movi(JIT_RA0 - (v->u.w - 8), uu.w);
-    }
+    else if (jit_arg_reg_p(v->u.w - 8))
+	jit_movi_d_w(JIT_RA0 - (v->u.w - 8), u);
     else {
+	jit_node_t	*node;
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_d(regno, u);
-	jit_stxi_d(v->u.w, JIT_FP, regno);
+	node = jit_stxi_d(v->u.w, JIT_FP, regno);
+	jit_link_alist(node);
 	jit_unget_reg(regno);
     }
     jit_dec_synth();
@@ -631,6 +672,7 @@ _jit_pushargr(jit_state_t *_jit, jit_int32_t u, jit_code_t code)
     else {
 	jit_stxi(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -652,6 +694,7 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u, jit_code_t code)
 	jit_stxi(_jitc->function->call.size, JIT_SP, regno);
 	jit_unget_reg(regno);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -674,6 +717,7 @@ _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
     else {
 	jit_stxi_f(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -700,6 +744,7 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
  	jit_stxi_f(_jitc->function->call.size, JIT_SP, regno);
 	jit_unget_reg(regno);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -722,6 +767,7 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
     else {
 	jit_stxi_d(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -748,6 +794,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
  	jit_stxi_d(_jitc->function->call.size, JIT_SP, regno);
 	jit_unget_reg(regno);
 	_jitc->function->call.size += sizeof(jit_word_t);
+	jit_check_frame();
     }
     jit_dec_synth();
 }
@@ -776,6 +823,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_check_frame();
     jit_inc_synth_w(finishr, r0);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
@@ -793,6 +841,7 @@ _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_check_frame();
     jit_inc_synth_w(finishi, (jit_word_t)i0);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
@@ -971,6 +1020,12 @@ _emit_code(jit_state_t *_jit)
 		name##r##type(rn(node->u.q.l), rn(node->u.q.h),		\
 			      rn(node->v.w), rn(node->w.w));		\
 		break
+#define case_rqr(name, type)						\
+	    case jit_code_##name##r##type:				\
+		name##r##type(rn(node->u.w), rn(node->v.q.l),		\
+			      rn(node->v.q.h), rn(node->w.w));		\
+	    case jit_code_##name##i##type:				\
+		break;
 #define case_rrw(name, type)						\
 	    case jit_code_##name##i##type:				\
 		name##i##type(rn(node->u.w), rn(node->v.w), node->w.w);	\
@@ -1114,10 +1169,46 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(lsh,);
 		case_rrr(rsh,);
 		case_rrw(rsh,);
+#define qlshr(r0, r1, r2, r3)	fallback_qlshr(r0, r1, r2, r3)
+#define qlshi(r0, r1, r2, i0)	fallback_qlshi(r0, r1, r2, i0)
+#define qlshr_u(r0, r1, r2, r3)	fallback_qlshr_u(r0, r1, r2, r3)
+#define qlshi_u(r0, r1, r2, i0)	fallback_qlshi_u(r0, r1, r2, i0)
+		case_rrrr(qlsh,);
+		case_rrrw(qlsh,);
+		case_rrrr(qlsh, _u);
+		case_rrrw(qlsh, _u);
 		case_rrr(rsh, _u);
 		case_rrw(rsh, _u);
+#define qrshr(r0, r1, r2, r3)	fallback_qrshr(r0, r1, r2, r3)
+#define qrshi(r0, r1, r2, i0)	fallback_qrshi(r0, r1, r2, i0)
+#define qrshr_u(r0, r1, r2, r3)	fallback_qrshr_u(r0, r1, r2, r3)
+#define qrshi_u(r0, r1, r2, i0)	fallback_qrshi_u(r0, r1, r2, i0)
+		case_rrrr(qrsh,);
+		case_rrrw(qrsh,);
+		case_rrrr(qrsh, _u);
+		case_rrrw(qrsh, _u);
+#define lrotr(r0,r1,r2)	fallback_lrotr(r0,r1,r2)
+#define lroti(r0,r1,i0)	fallback_lroti(r0,r1,i0)
+#define rrotr(r0,r1,r2)	fallback_rrotr(r0,r1,r2)
+#define rroti(r0,r1,i0)	fallback_rroti(r0,r1,i0)
+		case_rrr(lrot,);
+		case_rrw(lrot,);
+		case_rrr(rrot,);
+		case_rrw(rrot,);
 		case_rr(neg,);
 		case_rr(com,);
+#define clor(r0, r1)	fallback_clo(r0, r1)
+#define clzr(r0, r1)	fallback_clz(r0, r1)
+#define ctor(r0, r1)	fallback_cto(r0, r1)
+#define ctzr(r0, r1)	fallback_ctz(r0, r1)
+#define rbitr(r0, r1)	fallback_rbit(r0, r1)
+#define popcntr(r0, r1)	fallback_popcnt(r0, r1)
+		case_rr(clo,);
+		case_rr(clz,);
+		case_rr(cto,);
+		case_rr(ctz,);
+		case_rr(rbit,);
+		case_rr(popcnt,);
 		case_rrr(and,);
 		case_rrw(and,);
 		case_rrr(or,);
@@ -1156,6 +1247,18 @@ _emit_code(jit_state_t *_jit)
 		case_rrw(ldx, _ui);
 		case_rrr(ldx, _l);
 		case_rrw(ldx, _l);
+	    case jit_code_unldr:
+		unldr(rn(node->u.w), rn(node->v.w), node->w.w);
+		break;
+	    case jit_code_unldi:
+		unldi(rn(node->u.w), node->v.w, node->w.w);
+		break;
+	    case jit_code_unldr_u:
+		unldr_u(rn(node->u.w), rn(node->v.w), node->w.w);
+		break;
+	    case jit_code_unldi_u:
+		unldi_u(rn(node->u.w), node->v.w, node->w.w);
+		break;
 		case_rr(st, _c);
 		case_wr(st, _c);
 		case_rr(st, _s);
@@ -1172,12 +1275,33 @@ _emit_code(jit_state_t *_jit)
 		case_wrr(stx, _i);
 		case_rrr(stx, _l);
 		case_wrr(stx, _l);
+	    case jit_code_unstr:
+		unstr(rn(node->u.w), rn(node->v.w), node->w.w);
+		break;
+	    case jit_code_unsti:
+		unsti(node->u.w, rn(node->v.w), node->w.w);
+		break;
 		case_rr(hton, _us);
 		case_rr(hton, _ui);
 		case_rr(hton, _ul);
 		case_rr(bswap, _us);
 		case_rr(bswap, _ui);
 		case_rr(bswap, _ul);
+#define extr(r0, r1, i0, i1)	fallback_ext(r0, r1, i0, i1)
+#define extr_u(r0, r1, i0, i1)	fallback_ext_u(r0, r1, i0, i1)
+#define depr(r0, r1, i0, i1)	fallback_dep(r0, r1, i0, i1)
+	    case jit_code_extr:
+		extr(rn(node->u.w), rn(node->v.w), node->w.q.l, node->w.q.h);
+		break;
+	    case jit_code_extr_u:
+		extr_u(rn(node->u.w), rn(node->v.w), node->w.q.l, node->w.q.h);
+		break;
+	    case jit_code_depr:
+		depr(rn(node->u.w), rn(node->v.w), node->w.q.l, node->w.q.h);
+		break;
+	    case jit_code_depi:
+		depi(rn(node->u.w), node->v.w, node->w.q.l, node->w.q.h);
+		break;
 		case_rr(ext, _c);
 		case_rr(ext, _uc);
 		case_rr(ext, _s);
@@ -1284,15 +1408,31 @@ _emit_code(jit_state_t *_jit)
 		case_rr(abs, _f);
 		case_rr(neg, _f);
 		case_rr(sqrt, _f);
+		case_rqr(fma, _f);
+		case_rqr(fms, _f);
+		case_rqr(fnma, _f);
+		case_rqr(fnms, _f);
 		case_rr(ext, _f);
 		case_rr(ld, _f);
 		case_rw(ld, _f);
 		case_rrr(ldx, _f);
 		case_rrw(ldx, _f);
+	    case jit_code_unldr_x:
+		unldr_x(rn(node->u.w), rn(node->v.w), node->w.w);
+		break;
+	    case jit_code_unldi_x:
+		unldi_x(rn(node->u.w), node->v.w, node->w.w);
+		break;
 		case_rr(st, _f);
 		case_wr(st, _f);
 		case_rrr(stx, _f);
 		case_wrr(stx, _f);
+	    case jit_code_unstr_x:
+		unstr_x(rn(node->u.w), rn(node->v.w), node->w.w);
+		break;
+	    case jit_code_unsti_x:
+		unsti_x(node->u.w, rn(node->v.w), node->w.w);
+		break;
 		case_rr(mov, _f);
 	    case jit_code_movi_f:
 		assert_data(node);
@@ -1367,6 +1507,10 @@ _emit_code(jit_state_t *_jit)
 		case_rr(abs, _d);
 		case_rr(neg, _d);
 		case_rr(sqrt, _d);
+		case_rqr(fma, _d);
+		case_rqr(fms, _d);
+		case_rqr(fnma, _d);
+		case_rqr(fnms, _d);
 		case_rr(ext, _d);
 		case_rr(ld, _d);
 		case_rw(ld, _d);
@@ -1439,6 +1583,7 @@ _emit_code(jit_state_t *_jit)
 		case_brr(bunord, _d);
 		case_brd(bunord);
 	    case jit_code_jmpr:
+		jit_check_frame();
 		jmpr(rn(node->u.w));
 		break;
 	    case jit_code_jmpi:
@@ -1458,10 +1603,13 @@ _emit_code(jit_state_t *_jit)
 			patch(word, node);
 		    }
 		}
-		else
+		else {
+		    jit_check_frame();
 		    jmpi(node->u.w);
+		}
 		break;
 	    case jit_code_callr:
+		jit_check_frame();
 		callr(rn(node->u.w));
 		break;
 	    case jit_code_calli:
@@ -1481,8 +1629,10 @@ _emit_code(jit_state_t *_jit)
 			patch(word, node);
 		    }
 		}
-		else
+		else {
+		    jit_check_frame();
 		    calli(node->u.w);
+		}
 		break;
 	    case jit_code_prolog:
 		_jitc->function = _jitc->functions.ptr + node->w.w;
@@ -1492,8 +1642,11 @@ _emit_code(jit_state_t *_jit)
 #if DEVEL_DISASSEMBLER
 		undo.prevw = prevw;
 #endif
+		undo.const_offset = _jitc->consts.vector.offset;
 		undo.patch_offset = _jitc->patches.offset;
 	    restart_function:
+		compute_framesize();
+		patch_alist(0);
 		_jitc->again = 0;
 		prolog(node);
 		break;
@@ -1514,12 +1667,21 @@ _emit_code(jit_state_t *_jit)
 		     * the reason of the undo. */
 		    undo.func.self.aoff = _jitc->function->frame +
 			_jitc->function->self.aoff;
+		    undo.func.need_frame = _jitc->function->need_frame;
 		    jit_regset_set(&undo.func.regset, &_jitc->function->regset);
+		    /* allocar information also does not need to be undone */
+		    undo.func.aoffoff = _jitc->function->aoffoff;
+		    undo.func.allocar = _jitc->function->allocar;
+		    /* this will be recomputed but undo anyway to have it
+		     * better self documented.*/
+		    undo.func.need_stack = _jitc->function->need_stack;
 		    memcpy(_jitc->function, &undo.func, sizeof(undo.func));
 #if DEVEL_DISASSEMBLER
 		    prevw = undo.prevw;
 #endif
 		    _jitc->patches.offset = undo.patch_offset;
+		    _jitc->consts.vector.offset = undo.const_offset;
+		    patch_alist(1);
 		    goto restart_function;
 		}
 		/* remember label is defined */
@@ -1538,6 +1700,9 @@ _emit_code(jit_state_t *_jit)
 		assert_data(node);
 		movi_f_w(rn(node->u.w), node->v.f);
 		break;
+	    case jit_code_movi_w_f:
+		movi_w_f(rn(node->u.w), node->v.w);
+		break;
 	    case jit_code_movr_w_d:
 		movr_w_d(rn(node->u.w), rn(node->v.w));
 		break;
@@ -1548,6 +1713,9 @@ _emit_code(jit_state_t *_jit)
 		assert_data(node);
 		movi_d_w(rn(node->u.w), node->v.d);
 		break;
+	    case jit_code_movi_w_d:
+		movi_w_d(rn(node->u.w), node->v.w);
+		break;
 	    case jit_code_va_start:
 		vastart(rn(node->u.w));
 		break;
@@ -1605,6 +1773,77 @@ _emit_code(jit_state_t *_jit)
 	    case jit_code_retval_f:		case jit_code_retval_d:
 	    case jit_code_prepare:
 	    case jit_code_finishr:		case jit_code_finishi:
+	    case jit_code_negi_f:		case jit_code_absi_f:
+	    case jit_code_sqrti_f:		case jit_code_negi_d:
+	    case jit_code_absi_d:		case jit_code_sqrti_d:
+		break;
+	    case jit_code_negi:
+		negi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_comi:
+		comi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti_c:
+		exti_c(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti_uc:
+		exti_uc(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti_s:
+		exti_s(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti_us:
+		exti_us(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_bswapi_us:
+		bswapi_us(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_bswapi_ui:
+		bswapi_ui(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_htoni_us:
+		htoni_us(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_htoni_ui:
+		htoni_ui(rn(node->u.w), node->v.w);
+		break;
+#if __WORDSIZE == 64
+	    case jit_code_exti_i:
+		exti_i(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti_ui:
+		exti_ui(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_bswapi_ul:
+		bswapi_ul(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_htoni_ul:
+		htoni_ul(rn(node->u.w), node->v.w);
+		break;
+#endif
+	    case jit_code_cloi:
+		cloi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_clzi:
+		clzi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_ctoi:
+		ctoi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_ctzi:
+		ctzi(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_rbiti:
+		rbiti(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_popcnti:
+		popcnti(rn(node->u.w), node->v.w);
+		break;
+	    case jit_code_exti:
+		exti(rn(node->u.w), node->v.w, node->w.q.l, node->w.q.h);
+		break;
+	    case jit_code_exti_u:
+		exti_u(rn(node->u.w), node->v.w, node->w.q.l, node->w.q.h);
 		break;
 	    default:
 		abort();
@@ -1702,6 +1941,7 @@ _emit_code(jit_state_t *_jit)
 #define CODE				1
 #  include "jit_riscv-cpu.c"
 #  include "jit_riscv-fpu.c"
+#  include "jit_fallback.c"
 #undef CODE
 
 static void
@@ -1849,6 +2089,30 @@ _emit_stxi_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
     stxi_d(i0, rn(r0), rn(r1));
 }
 
+#if __WORDSIZE != 64
+# error "only 64 bit ports tested"
+#endif
+static void
+_compute_framesize(jit_state_t *_jit)
+{
+    jit_int32_t		reg;
+    _jitc->framesize = 16;	/* ra+fp */
+    for (reg = 0; reg < jit_size(iregs); reg++)
+	if (jit_regset_tstbit(&_jitc->function->regset, iregs[reg]))
+	    _jitc->framesize += sizeof(jit_word_t);
+
+    for (reg = 0; reg < jit_size(fregs); reg++)
+	if (jit_regset_tstbit(&_jitc->function->regset, fregs[reg]))
+	    _jitc->framesize += sizeof(jit_float64_t);
+
+    /* Space to store variadic arguments */
+    if (_jitc->function->self.call & jit_call_varargs)
+	_jitc->framesize += (8 - _jitc->function->vagp) * 8;
+
+    /* Make sure functions called have a 16 byte aligned stack */
+    _jitc->framesize = (_jitc->framesize + 15) & -16;
+}
+
 static void
 _patch(jit_state_t *_jit, jit_word_t instr, jit_node_t *node)
 {
