@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2022  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2023  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -65,14 +65,16 @@ typedef union {
     struct {	jit_uint32_t _: 13;	jit_uint32_t b: 5;	} rs1;
     struct {	jit_uint32_t _: 18;	jit_uint32_t b: 1;	} i;
     struct {	jit_uint32_t _: 18;	jit_uint32_t b: 9;	} opf;
+    struct {	jit_uint32_t _: 18;	jit_uint32_t b: 5;	} rs3;
     struct {	jit_uint32_t _: 19;	jit_uint32_t b: 1;	} x;
     struct {	jit_uint32_t _: 19;	jit_uint32_t b: 8;	} asi;
     struct {	jit_uint32_t _: 19;	jit_uint32_t b: 6;	} res;
     struct {	jit_uint32_t _: 19;	jit_uint32_t b: 13;	} simm13;
     struct {	jit_uint32_t _: 20;	jit_uint32_t b: 7;	} asix;
     struct {	jit_uint32_t _: 20;	jit_uint32_t b: 6;	} asis;
-    struct {	jit_uint32_t _: 26;	jit_uint32_t b: 6;	} shim;
+    struct {	jit_uint32_t _: 23;	jit_uint32_t b: 4;	} op5;
     struct {	jit_uint32_t _: 25;	jit_uint32_t b: 7;	} imm7;
+    struct {	jit_uint32_t _: 26;	jit_uint32_t b: 6;	} shim;
     struct {	jit_uint32_t _: 27;	jit_uint32_t b: 5;	} rs2;
     jit_int32_t							  v;
 } jit_instr_t;
@@ -99,7 +101,7 @@ _f2bp(jit_state_t*,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t,
 #  define f3r(op, rd, op3, rs1, rs2)	_f3r(_jit, op, rd, op3, rs1, rs2)
 static void _f3r(jit_state_t*,
 		 jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t);
-#  if __WORDSIZE == 64
+#  if __WORDSIZE == 64 || CHECK_LZCNT
 #  define f3ri(op, rd, op3, rs1, rs2)	_f3ri(_jit, op, rd, op3, rs1, rs2)
 static void _f3ri(jit_state_t*,
 		  jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t);
@@ -120,6 +122,11 @@ static void _f3t(jit_state_t*,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t)
 static void _f3a(jit_state_t*,jit_int32_t,
 		 jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t)
     maybe_unused;
+#  define f2c1(op,rd,op3,rs1,opf,rs2)	_f2c1(_jit,op,rd,op3,rs1,opf,rs2)
+static void
+_f2c1(jit_state_t*,jit_int32_t, jit_int32_t,
+      jit_int32_t,jit_int32_t,jit_int32_t,jit_int32_t)
+    maybe_unused;
 #  define LDSB(rs1, rs2, rd)		f3r(3, rd, 9, rs1, rs2)
 #  define LDSBI(rs1, imm, rd)		f3i(3, rd, 9, rs1, imm)
 #  define LDSH(rs1, rs2, rd)		f3r(3, rd, 10, rs1, rs2)
@@ -238,7 +245,7 @@ static void _f3a(jit_state_t*,jit_int32_t,
 #  define SRLI(rs1, imm, rd)		f3i(2, rd, 38, rs1, imm)
 #  define SRA(rs1, rs2, rd)		f3r(2, rd, 39, rs1, rs2)
 #  define SRAI(rs1, imm, rd)		f3i(2, rd, 39, rs1, imm)
-#  if __WORDSIZE == 64
+#  if __WORDSIZE == 64 || CHECK_LZCNT
 #    define SLLX(rs1, rs2, rd)		f3rx(2, rd, 37, rs1, rs2)
 #    define SLLXI(rs1, imm, rd)		f3s(2, rd, 37, rs1, imm)
 #    define SRLX(rs1, rs2, rd)		f3rx(2, rd, 38, rs1, rs2)
@@ -545,6 +552,7 @@ static void _f3a(jit_state_t*,jit_int32_t,
 #  define UNIMP(imm)			f2r(0, 0, 0, imm)
 #  define FLUSH(rs1, rs2)		f3r(2, 0, 59, rs1, rs2)
 #  define FLUSHI(rs1, im)		f3i(2, 0, 59, rs1, imm)
+#  define LZCNT(rs2, rd)		f2c1(2, rd, 54, 0, 23, rs2)
 #  define nop(i0)			_nop(_jit, i0)
 static void _nop(jit_state_t*, jit_int32_t);
 #  define movr(r0, r1)			_movr(_jit, r0, r1)
@@ -567,6 +575,14 @@ static void _casx(jit_state_t *_jit,jit_int32_t,jit_int32_t,
 #define casi(r0, i0, r1, r2)		casx(r0, _NOREG, r1, r2, i0)
 #  define comr(r0, r1)			XNOR(r1, 0, r0)
 #  define negr(r0, r1)			NEG(r1, r0)
+#  define clor(r0, r1)			_clor(_jit, r0, r1)
+static void _clor(jit_state_t*, jit_int32_t, jit_int32_t);
+#  define clzr(r0, r1)			_clzr(_jit, r0, r1)
+static void _clzr(jit_state_t*, jit_int32_t, jit_int32_t);
+#  define ctor(r0, r1)			_ctor(_jit, r0, r1)
+static void _ctor(jit_state_t*, jit_int32_t, jit_int32_t);
+#  define ctzr(r0, r1)			_ctzr(_jit, r0, r1)
+static void _ctzr(jit_state_t*, jit_int32_t, jit_int32_t);
 #  define addr(r0, r1, r2)		ADD(r1, r2, r0)
 #  define addi(r0, r1, i0)		_addi(_jit, r0, r1, i0)
 static void _addi(jit_state_t*, jit_int32_t, jit_int32_t, jit_word_t);
@@ -1182,6 +1198,26 @@ _f1(jit_state_t *_jit, jit_int32_t op, jit_int32_t disp30)
     ii(v.v);
 }
 
+static void
+_f2c1(jit_state_t *_jit, jit_int32_t op, jit_int32_t rd,
+      jit_int32_t op3, jit_int32_t rs1, jit_int32_t opf, jit_int32_t rs2)
+{
+    jit_instr_t		v;
+    assert(!(op  & 0xfffffffc));
+    assert(!(rd  & 0xffffffe0));
+    assert(!(res & 0xffffffc0));
+    assert(!(rs1 & 0xffffffe0));
+    assert(!(opf & 0xfffffe00));
+    assert(!(rs2 & 0xfffffe00));
+    v.op.b = op;
+    v.rd.b = rd;
+    v.op3.b = op3;
+    v.rs1.b = rs1;
+    v.opf.b = opf;
+    v.rs2.b = rs2;
+    ii(v.v);
+}
+
 static void
 _nop(jit_state_t *_jit, jit_int32_t i0)
 {
@@ -1296,6 +1332,68 @@ _casx(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1,
         jit_unget_reg(r1_reg);
 }
 
+static void
+_clor(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    if (jit_cpu.lzcnt) {
+	comr(r0, r1);
+	clzr(r0, r0);
+    }
+    else
+	fallback_clo(r0, r1);
+}
+
+static void
+_clzr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+#  if CHECK_LZCNT
+    if (jit_cpu.lzcnt) {
+#    if __WORDSIZE == 32
+	jit_word_t		w;
+	SLLXI(r1, 32, r0);
+	LZCNT(r0, r0);
+	w = blei(_jit->pc.w, r0, 31);
+	rshi(r0, r0, 1);	/* r0 is 64 */
+	patch_at(w, _jit->pc.w);
+#    else
+	LZCNT(r1, r0);
+#    endif
+    }
+    else
+#  endif
+	fallback_clz(r0, r1);
+}
+
+static void
+_ctor(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    if (jit_cpu.lzcnt) {
+	comr(r0, r1);
+	ctzr(r0, r0);
+    }
+    else
+	fallback_cto(r0, r1);
+}
+
+static void
+_ctzr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    jit_int32_t		t0, t1;
+    if (jit_cpu.lzcnt) {
+	t0 = jit_get_reg(jit_class_gpr);
+	t1 = jit_get_reg(jit_class_gpr);
+	negr(rn(t0), r1);
+	andr(rn(t0), rn(t0), r1);
+	clzr(r0, rn(t0));
+	xori(rn(t1), r0, __WORDSIZE - 1);
+	movnr(r0, rn(t1), rn(t0));
+	jit_unget_reg(t0);
+	jit_unget_reg(t1);
+    }
+    else
+	fallback_ctz(r0, r1);
+}
+
 static void
 _addi(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1, jit_word_t i0)
 {
@@ -2307,7 +2405,11 @@ _bw(jit_state_t *_jit, jit_int32_t cc,
     if (s13_p(i1)) {
 	CMPI(r0, i1);
 	w = _jit->pc.w;
+#  if __WORDSIZE == 32
 	B(cc, (i0 - w) >> 2);
+#  else
+	BP(cc, (i0 - w) >> 2);
+#  endif
 	NOP();
     }
     else {
@@ -2551,24 +2653,24 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
 {
     if (_jitc->function->assume_frame)
 	return;
-    /* (most) other backends do not save incoming arguments, so,
-     * only save locals here */
+    if (_jitc->function->allocar)
+	subi(_SP_REGNO, _FP_REGNO, _jitc->function->stack);
     if (jit_regset_tstbit(&_jitc->function->regset, _L0))
-	ldxi(_L0_REGNO, _FP_REGNO, _jitc->function->stack + OFF(0));
+	ldxi(_L0_REGNO, _SP_REGNO, _jitc->function->stack + OFF(0));
     if (jit_regset_tstbit(&_jitc->function->regset, _L1))
-	ldxi(_L1_REGNO, _FP_REGNO, _jitc->function->stack + OFF(1));
+	ldxi(_L1_REGNO, _SP_REGNO, _jitc->function->stack + OFF(1));
     if (jit_regset_tstbit(&_jitc->function->regset, _L2))
-	ldxi(_L2_REGNO, _FP_REGNO, _jitc->function->stack + OFF(2));
+	ldxi(_L2_REGNO, _SP_REGNO, _jitc->function->stack + OFF(2));
     if (jit_regset_tstbit(&_jitc->function->regset, _L3))
-	ldxi(_L3_REGNO, _FP_REGNO, _jitc->function->stack + OFF(3));
+	ldxi(_L3_REGNO, _SP_REGNO, _jitc->function->stack + OFF(3));
     if (jit_regset_tstbit(&_jitc->function->regset, _L4))
-	ldxi(_L4_REGNO, _FP_REGNO, _jitc->function->stack + OFF(4));
+	ldxi(_L4_REGNO, _SP_REGNO, _jitc->function->stack + OFF(4));
     if (jit_regset_tstbit(&_jitc->function->regset, _L5))
-	ldxi(_L5_REGNO, _FP_REGNO, _jitc->function->stack + OFF(5));
+	ldxi(_L5_REGNO, _SP_REGNO, _jitc->function->stack + OFF(5));
     if (jit_regset_tstbit(&_jitc->function->regset, _L6))
-	ldxi(_L6_REGNO, _FP_REGNO, _jitc->function->stack + OFF(6));
+	ldxi(_L6_REGNO, _SP_REGNO, _jitc->function->stack + OFF(6));
     if (jit_regset_tstbit(&_jitc->function->regset, _L7))
-	ldxi(_L7_REGNO, _FP_REGNO, _jitc->function->stack + OFF(7));
+	ldxi(_L7_REGNO, _SP_REGNO, _jitc->function->stack + OFF(7));
     RESTOREI(0, 0, 0);
     RETL();
     NOP();
