@@ -40,9 +40,14 @@ main(int argc, char *argv[])
     fclose(fp);
 
     max = 0;
-    for (offset = 0; offset < jit_code_last_code; offset++)
+    for (offset = 0; offset < jit_code_last_code; offset++) {
+#if defined(__ia64__)
+	if (_szs[offset] > 16)
+	    _szs[offset] = _szs[offset] / 3 + 16 & -16;
+#endif
 	if (max < _szs[offset])
 	    max = _szs[offset];
+    }
 
     if ((fp = fopen(JIT_SIZE_PATH, "w")) == NULL)
 	exit(-1);
@@ -68,14 +73,6 @@ main(int argc, char *argv[])
 #  else
     fprintf(fp, "#if !defined(__ARM_PCS_VFP)\n");
 #  endif
-#elif defined(__mips__)
-#  if __WORDSIZE == 32
-#    if NEW_ABI
-    fprintf(fp, "#if NEW_ABI\n");
-#    else
-    fprintf(fp, "#if !NEW_ABI\n");
-#    endif
-#  endif
 #elif defined(__powerpc__)
     fprintf(fp, "#if defined(__powerpc__)\n");
     fprintf(fp, "#if __BYTE_ORDER == %s\n",
@@ -94,10 +91,6 @@ main(int argc, char *argv[])
 	fprintf(fp, "    %d,	/* %s */\n", _szs[offset], code_name[offset]);
 #if defined(__arm__)
     fprintf(fp, "#endif /* __ARM_PCS_VFP */\n");
-#elif defined(__mips__)
-#  if __WORDSIZE == 32
-    fprintf(fp, "#endif /* NEW_ABI */\n");
-#  endif
 #elif defined(__powerpc__)
 #  if __WORDSIZE == 32
     fprintf(fp, "#endif /* "
